start.time  <-  Sys.time()

#########################
# LIBRARY AND FUNCTIONS #
#########################

# Use the following line to load the Snakemake object to manually rerun this script (e.g., for debugging purposes)
# Replace {outputFolder} correspondingly.
# snakemake = readRDS("{outputFolder}/LOGS_AND_BENCHMARKS/2.DESeqPeaks.R.rds")


library("checkmate")
assertClass(snakemake, "Snakemake")
assertDirectoryExists(snakemake@config$par_general$scriptsDir)
source(paste0(snakemake@config$par_general$scriptsDir, "/functions.R"))

########################################################################
# SAVE SNAKEMAKE S4 OBJECT THAT IS PASSED ALONG FOR DEBUGGING PURPOSES #
########################################################################

createDebugFile(snakemake)

initFunctionsScript(packagesReq = NULL, minRVersion = "3.1.0", warningsLevel = 1, disableScientificNotation = TRUE)
checkAndLoadPackages(c("tidyverse", "futile.logger",  "checkmate", "tools", "methods", "DESeq2"), verbose = FALSE)



###################
#### PARAMETERS ###
###################

par.l = list()

par.l$verbose = TRUE
par.l$log_minlevel = "INFO"

#####################
# VERIFY PARAMETERS #
#####################

checkAndLogWarningsAndErrors(snakemake, checkClass(snakemake, "Snakemake"))

## INPUT ##
checkAndLogWarningsAndErrors(snakemake@input, checkList(snakemake@input, min.len = 3))
checkAndLogWarningsAndErrors(snakemake@input, checkSubset(names(snakemake@input), c("", "metadata", "rawCounts", "rawCountsMerged")))


par.l$file_metadata = snakemake@input$metadata
par.l$file_rawCounts = snakemake@input$rawCounts
par.l$file_rawCountsMerged = snakemake@input$rawCountsMerged

checkAndLogWarningsAndErrors(par.l$file_metadata, checkFileExists(par.l$file_metadata, access = "r"))
checkAndLogWarningsAndErrors(par.l$file_rawCounts, checkFileExists(par.l$file_rawCounts, access = "r"))
checkAndLogWarningsAndErrors(par.l$file_rawCountsMerged, checkFileExists(par.l$file_rawCountsMerged, access = "r"))


## OUTPUT ##
checkAndLogWarningsAndErrors(snakemake@output, checkList(snakemake@output, min.len = 2))
checkAndLogWarningsAndErrors(names(snakemake@output), checkSubset(names(snakemake@output), c("", "pdf", "pdfMerged")))

par.l$file_output_pdf       = snakemake@output$pdf
par.l$file_output_pdfMerged = snakemake@output$pdfMerged


## PARAMS ##
checkAndLogWarningsAndErrors(snakemake@params, checkList(snakemake@params, min.len = 4))
checkAndLogWarningsAndErrors(names(snakemake@params), checkSubset(names(snakemake@params), c("", "binSize", "filterChr", "splitByGCCorrection", "minNoCountsRows")))

par.l$binSize = as.integer(snakemake@params$binSize)
checkAndLogWarningsAndErrors(par.l$binSize, checkIntegerish(par.l$binSize))

par.l$filterChr = as.logical(snakemake@params$filterChr)
checkAndLogWarningsAndErrors(par.l$filterChr, checkFlag(par.l$filterChr))

par.l$splitByGCCorrection = as.logical(snakemake@params$splitByGCCorrection)
checkAndLogWarningsAndErrors(par.l$splitByGCCorrection, checkFlag(par.l$splitByGCCorrection))

par.l$minNoCountsRows = as.integer(snakemake@params$minNoCountsRows)
checkAndLogWarningsAndErrors(par.l$minNoCountsRows, checkIntegerish(par.l$minNoCountsRows))

## LOG ##
checkAndLogWarningsAndErrors(snakemake@log, checkList(snakemake@log, min.len = 1))
par.l$file_log = snakemake@log[[1]]


allDirs = c(dirname(par.l$file_output_pdfMerged), 
            dirname(par.l$file_output_pdf),
            dirname(par.l$file_log)
)

testExistanceAndCreateDirectoriesRecursively(allDirs)


######################
# FINAL PREPARATIONS #
######################
startLogger(par.l$file_log, par.l$log_minlevel,  removeOldLog = TRUE)
printParametersLog(par.l)

readCounts.l = list(all = par.l$file_rawCounts, pooled = par.l$file_rawCountsMerged)
output.l     = list(all = par.l$file_output_pdf, pooled = par.l$file_output_pdfMerged)

for (typeCur in c("all", "pooled")) {
  
  flog.info(paste0("Generate PCA plots for ", typeCur, " samples"))
  
  readCountsCur  = readCounts.l[[typeCur]]
  outputCur      = output.l[[typeCur]]
  
  doPCAPlot(file_readCounts = readCountsCur, 
            file_metadata = par.l$file_metadata, 
            file_output = outputCur, 
            file_log = par.l$file_log, 
            metadataToInclude = c("sampleName", "individual"), 
            type = typeCur,
            binSize = par.l$binSize, 
            filterChr = par.l$filterChr, 
            splitByGCCorrection = par.l$splitByGCCorrection, 
            minNoCountsRows = par.l$minNoCountsRows)
}

