# Almost all of these functions are originally described in source("/g/scb2/zaugg/zaugg_shared/scripts/Christian/src/R/functionsCollection.R")
# and have been copied from there 


.prettyNum <- function(number, verbose = TRUE) {
    prettyNum(number, big.mark = ",", scientific = FALSE)
}


startLogger <- function(logfile, level, removeOldLog = TRUE, appenderName = "consoleAndFile", verbose = FALSE) {
    
    checkAndLoadPackages(c("futile.logger"), verbose = verbose)
    
    assertSubset(level, c("TRACE", "DEBUG", "INFO", "WARN", "ERROR", "FATAL"))
    assertFlag(removeOldLog)
    assertSubset(appenderName, c("console", "file", "consoleAndFile"))
    assertFlag(verbose)
    
    if (appenderName != "console") {
        assertDirectory(dirname(logfile), access = "w")
        if (file.exists(logfile)) {
            file.remove(logfile)
        }
    }
    
    # LEVELS: TRACE, DEBUG, INFO, WARN, ERROR, FATAL
    invisible(flog.threshold(level))
    
    
    if (appenderName == "console") {
        invisible(flog.appender(appender.console()))
    } else if (appenderName == "file")  {
        invisible(flog.appender(appender.file(file = logfile)))
    } else {
        invisible(flog.appender(appender.tee(file = logfile)))
    }
    
    
}

printParametersLog <- function(par.l, verbose = FALSE) {
    
    checkAndLoadPackages(c("futile.logger"), verbose = verbose)  
    assertList(par.l)
    flog.info(paste0("PARAMETERS:"))
    for (parCur in names(par.l)) {
        
        flog.info(paste0(" ", parCur, "=",  paste0(par.l[[parCur]], collapse = ",")))
        
    }
}


clearOpenDevices <- function() {
    
    while (length(dev.list()) > 0) {
        dev.off()
    }
}


printMultipleGraphsPerPage <- function(plots.l, nCol = 1, nRow = 1, pdfFile = NULL, height = NULL, width = NULL, verbose = FALSE) {
    
    checkAndLoadPackages(c("grDevices", "gridExtra"), verbose = verbose)  
    
    assertList(plots.l, min.len = 1)
    assertInt(nCol, lower = 1)
    assertInt(nRow, lower = 1)
    assert(checkNull(pdfFile), checkCharacter(pdfFile))
    if (!testNull(pdfFile)) assertDirectory(dirname(pdfFile), access = "r")
    assert(checkNull(height), checkInt(height, lower = 1))
    assert(checkNull(width), checkInt(width, lower = 1))
    
    if (testNull(height)) {
        height = 7
    }
    
    if (testNull(width)) {
        width = 7
    }
    
    
    for (i in seq_len(length(plots.l))) {
        assertClass(plots.l[[i]], classes = c("ggplot", "gg"))
    }
    
    
    nPlotsPerPage =  nCol * nRow
    
    plotsNew.l = list()
    
    if (!testNull(pdfFile))  {
      clearOpenDevices()
      pdf(pdfFile, height = height, width = width)
    }
    index = 0
    
    for (indexAll in 1:length(plots.l)) {
        
        index = index + 1
        plotsNew.l[[index]] = plots.l[[indexAll]]
        
        # Print another page
        if (index %% nPlotsPerPage == 0) { ## print 8 plots on a page
            suppressMessages(print(do.call(grid.arrange,  c(plotsNew.l, list(ncol = nCol, nrow = nRow)))))
            plotsNew.l = list() # reset plot 
            index = 0 # reset index
        }
        
    }
    
    # Print the remainding plots in case they don't perfectly fit with the layout
    if (length(plotsNew.l) != 0) { 
        suppressMessages(print(do.call(grid.arrange,  c(plotsNew.l, list(ncol = nCol, nrow = nRow)))))
    }
    
    if (!testNull(pdfFile)) 
        dev.off()
    
    cat("Finished writing plots to file ", pdfFile, "\n")
}


initFunctionsScript <- function(packagesReq = NULL, minRVersion = "3.1.0", warningsLevel = 1, disableScientificNotation = TRUE, verbose = TRUE) {
    
    checkAndLoadPackages("checkmate", verbose = verbose)
    assert(checkNull(packagesReq), checkCharacter(packagesReq, min.len = 1, min.chars = 1))
    assertCharacter(minRVersion, len = 1)
    assertInt(warningsLevel, lower = 0, upper = 2)
    assertFlag(disableScientificNotation)
    assertFlag(verbose)
    
    clearOpenDevices()
    
    
    # No annoying strings as factors by default
    options(stringsAsFactors = FALSE)
    
    # Print warnings as they occur
    options(warn = warningsLevel)
    
    # Just print 50 lines instead of 99999
    options(max.print = 200)
    
    # Disable scientific notation
    if (disableScientificNotation) options(scipen = 999)
    
    
    # We need at least R version 3.1.0 to continue
    stopifnot(getRversion() >= minRVersion)

    
    .detachAllPackages()
    
    checkAndLoadPackages(packagesReq, verbose = verbose)
    
}

checkAndLoadPackages <- function(packages, verbose = TRUE) {
    
    .checkAndInstallMissingPackages(packages, verbose = verbose)
    
    for (packageCur in packages) {
        library(packageCur, character.only = TRUE)
    }
    
}


.checkAndInstallMissingPackages <- function(packages.vec, verbose = TRUE, cranMirror = "http://cran.uni-muenster.de/") {
    
    if (verbose) cat("Trying to automatically install missing packages. If this fails, install them manually...\n")
    
    packagesToInstall = setdiff(packages.vec, rownames(installed.packages()))
    
    
    if (length(packagesToInstall) > 0) {
        if (verbose) cat("Could not find the following packages: ", paste( packagesToInstall , collapse = ", "), "\n")
        install.packages(packagesToInstall, repos = cranMirror)  
        
        source("http://bioconductor.org/biocLite.R")
        for (packageCur in packagesToInstall) {
            biocLite(packageCur, suppressUpdates = TRUE)
        }
    } else {
        if (verbose) cat("All packages are already installed\n")
    }
    
}


createDebugFile <- function(snakemake) {
  
  checkAndLoadPackages(c("checkmate", "tools", "futile.logger"), verbose = FALSE)
  
  if (!testClass(snakemake, "Snakemake")) {
    flog.warn(paste0("Could not find snakemake object, therefore not saving anyting."))
    return()
  }
  
  logfile = snakemake@log[[1]][1]
  
  assertCharacter(logfile, any.missing = FALSE)
  
  filename = paste0(tools::file_path_sans_ext(logfile), ".rds")
  
  # flog.info(paste0("Saved Snakemake object for manually rerunning the R script to ", filename))
  
  saveRDS(snakemake, filename)
  
}


testExistanceAndCreateDirectoriesRecursively <- function(directories) {
  
  for (dirname in unique(directories)) {
    
    if (!testDirectoryExists(dirname)) {
      dir.create(dirname, recursive = TRUE)
    } else {
      assertDirectoryExists(dirname, access = "w")
    }
    
  }
}


.printExecutionTime <- function(startTime) {
  
  endTime  <-  Sys.time()
  flog.info(paste0("Script finished sucessfully. Execution time: ", round(endTime - startTime, 1), " ", units(endTime - startTime)))
}

.detachAllPackages <- function() {
    
    basic.packages <- c("package:stats","package:graphics","package:grDevices","package:utils","package:datasets","package:methods","package:base")
    
    package.list <- search()[ifelse(unlist(gregexpr("package:",search())) == 1,TRUE,FALSE)]
    
    package.list <- setdiff(package.list,basic.packages)
    
    if (length(package.list) > 0)  for (package in package.list) detach(package, character.only = TRUE)
    
}


createFileList <- function(directory, pattern, recursive = FALSE, ignoreCase = FALSE, verbose = TRUE) {
    
    assertCharacter(directory, min.chars = 1, any.missing = FALSE, len = 1)
    assertCharacter(pattern, min.chars = 1, any.missing = FALSE, len = 1)
    assertFlag(recursive)  
    assertFlag(ignoreCase)  
    assertFlag(verbose)    
    
    assertDirectoryExists(directory)
    
    # Multiple patterns are now supported, integrate over them
    patternAll = strsplit(pattern, ",")[[1]]
    assertCharacter(patternAll, min.len = 1)
    
    if (verbose) cat("Found ", length(patternAll), " distinct pattern(s) in pattern string.\n")
    
    nFilesToProcessTotal = 0
    filesToProcess.vec = c()
    
    for (patternCur in patternAll) {
        
        # Replace wildcards by functioning patterns (such as .)
        patternMod = glob2rx(patternCur)
        
        # Remove anchoring at beginning and end
        patternMod = substr(patternMod, 2, nchar(patternMod) - 1)
        
        filesToProcessCur.vec = list.files(path = directory, pattern = patternMod, full.names = TRUE, recursive = recursive, ignore.case = ignoreCase)
        filesToProcess.vec = c(filesToProcess.vec, filesToProcessCur.vec)
        
        if (verbose) cat("Search for files with pattern \"", patternCur, "\" in directory ", directory, " (case insensitive:", ignoreCase, ")\n", sep ="")
        
        nFilesToProcessTotal = nFilesToProcessTotal + length(filesToProcessCur.vec)
    }
    
    
    
    if (nFilesToProcessTotal == 0) {
        stop(paste0("No files to process in folder ", directory, " that fulfill the desired criteria (", patternCur, ")."))
    } else {
        
        if (verbose) cat("The following", nFilesToProcessTotal, "files were found:\n", paste0(filesToProcess.vec, collapse = "\n "))
    }
    
    if (verbose) cat("\n")
    
    filesToProcess.vec
    
}


.getThemeForGGPlot <- function(verticalLines = FALSE, horizontalLines = FALSE, gridColor = "gray", linetype = "dotted", lineSize = 0.5, centerTitle = TRUE, verbose = FALSE) {
    
    checkAndLoadPackages(c("ggplot2"), verbose = verbose)  
    assertFlag(horizontalLines)
    assertFlag(verticalLines)
    assertFlag(verbose)
    assertNumber(lineSize, lower = 0)
    assertSubset(linetype, c("blank", "solid", "dashed", "dotted", "dotdash", "longdash", "twodash"))
    assertFlag(centerTitle)
    
    gridX = element_blank()
    gridY = element_blank()
    
    if (verticalLines) {
        gridX = element_line(colour = gridColor, linetype = linetype, size = lineSize)
    }
    if (horizontalLines) {
        gridY = element_line(colour = gridColor, linetype = linetype, size = lineSize)
    }
    
    
    theme(panel.grid.minor.y = gridY, panel.grid.major.y = gridY,
          panel.grid.minor.x = gridX, panel.grid.major.x = gridX) + 
        theme_bw() + 
        theme(plot.title = element_text(hjust =  ifelse(centerTitle, 0.5, 0)))
    
}


#' @import checkmate
generateDefaultReadFlags <- function(pairedEndReads = TRUE, verbose = TRUE) {
    
    assertFlag(pairedEndReads)
    assertFlag(verbose)
    
    par.l = list(  
        
        "readFlag_isPaired" = TRUE, 
        "readFlag_isProperPair" = TRUE ,
        "readFlag_isUnmappedQuery" = FALSE, 
        "readFlag_hasUnmappedMate" = FALSE, 
        "readFlag_isMinusStrand" = NA, 
        "readFlag_isMateMinusStrand" = NA,
        "readFlag_isFirstMateRead" = NA, 
        "readFlag_isSecondMateRead" = NA, 
        "readFlag_isNotPrimaryRead" = FALSE,
        "readFlag_isNotPassingQualityControls" = FALSE, 
        "readFlag_isDuplicate" =  FALSE
    )
    
    if (!pairedEndReads) {
        
        par.l$readFlag_isPaired = NA 
        par.l$readFlag_isProperPair = NA 
        par.l$readFlag_isUnmappedQuery = NA 
        par.l$readFlag_hasUnmappedMate = NA 
        par.l$readFlag_isMateMinusStrand = NA 
        par.l$readFlag_isFirstMateRead = NA 
        par.l$readFlag_isSecondMateRead = NA 
        par.l$readFlag_isNotPrimaryRead = NA 
        
        
    }
    
    .constructScanBamFlags(par.l, verbose = verbose)
}


.constructScanBamFlags <- function(par.l, verbose = TRUE) {
    
    checkAndLoadPackages(c("Rsamtools"), verbose = verbose)
    
    namesReqElems = c("readFlag_isPaired", "readFlag_isProperPair", "readFlag_isUnmappedQuery", "readFlag_hasUnmappedMate", 
                      "readFlag_isMinusStrand", "readFlag_isMateMinusStrand", "readFlag_isFirstMateRead", 
                      "readFlag_isSecondMateRead", "readFlag_isNotPrimaryRead", "readFlag_isNotPassingQualityControls")
    
    assertList(par.l, min.len = length(namesReqElems))
    
    assertSubset(namesReqElems, names(par.l))
    
    flags = scanBamFlag(isPaired = par.l$readFlag_isPaired, 
                        isProperPair = par.l$readFlag_isProperPair, 
                        isUnmappedQuery = par.l$readFlag_isUnmappedQuery, 
                        hasUnmappedMate = par.l$readFlag_hasUnmappedMate, 
                        isMinusStrand = par.l$readFlag_isMinusStrand, 
                        isMateMinusStrand = par.l$readFlag_isMateMinusStrand, 
                        isFirstMateRead = par.l$readFlag_isFirstMateRead, 
                        isSecondMateRead = par.l$readFlag_isSecondMateRead, 
                        isSecondaryAlignment = par.l$readFlag_isNotPrimaryRead, 
                        isNotPassingQualityControls = par.l$readFlag_isNotPassingQualityControls, 
                        isDuplicate = par.l$readFlag_isDuplicate
    )
    
    flags
    
}



#TODO: Set 0 counts to 0.1 or 1 to allow enrichment values
#Soll man paired end reads doppelt zählen? Strange effect with even and odd numbers
# readFlags =  generateDefaultReadFlags(pairedEndReads = TRUE)

#  annotationFile = par.l$annotationFile
#  BAMFile = fileCur
#  thresholdWithinTSS = par.l$TSSEnrichment_withinThreshold
# thresholdOutsideTSS = par.l$TSSEnrichment_outsideThreshold
#  readFlags = generateDefaultReadFlags(pairedEndReads = par.l$pairedEnd)
#  geneTypesToKeep = par.l$TSSEnrichment_geneTypesToKeep
#  verbose = par.l$verbose

checkEnrichmentOverTSS <- function(annotationFile, BAMFile, thresholdWithinTSS = 4000, thresholdOutsideTSS = 1000, 
                                   readFlags, geneTypesToKeep = c("protein_coding"), verbose = TRUE) {
  
  checkAndLoadPackages(c("tools", "GenomicRanges"), verbose = verbose)
  
  assertFile(annotationFile, access = "r")
  assertFile(BAMFile, access = "r")
  assertSubset(tolower(file_ext(BAMFile)), c("bam"))
  assertInt(thresholdWithinTSS, lower = 1)
  assertInt(thresholdOutsideTSS, lower = 1)
  assert(checkNull(readFlags), checkIntegerish(readFlags, len = 2, lower = 0))
  assertCharacter(geneTypesToKeep, min.chars = 1, min.len = 1)
  assertFlag(verbose) 
  
  if (testNull(readFlags)) {
    readFlags = scanBamFlag()
  }
  
  
  flog.info(paste0("Check enrichment over TSS for file: ", BAMFile))
  flog.info(paste("thresholdWithinTSS=", thresholdWithinTSS, ", thresholdOutsideTSS=", thresholdOutsideTSS))
  
  flog.info(paste("Retrieving TSS annotation from file", annotationFile, sep = " "))
  filename_TSS = paste0(annotationFile, ".TSS_nonOverlapping", (thresholdWithinTSS + thresholdOutsideTSS), ".bed")
  
  if (!file.exists(filename_TSS)) {
    
    flog.info(paste0("Could not find file", filename_TSS, ", calculating TSS anew...", sep = " "))
    
    # Read only particular columns(positions and type) 
    colsToRead = c("character", "NULL", "character", "integer", "integer", "NULL", "character", "NULL", "character")
    
    genomeAnnotation.df =  read.table(annotationFile, colClasses = colsToRead, skip = 5, sep = "\t")
    colnames( genomeAnnotation.df) = c("chr", "type", "start", "end", "strand", "annot")
    
    # How to define a TSS? Current solution:
    # The first codon of the first exon is approximately the transcription start site. The translation start site(i.e. the start of the CDS) is located somewhere within the transcript, but doesn't necessarily have to be in the first exon at all.
    # The first nucleotide of the first exon is a good approximation. The main issue with it is that the exact TSS can shift around a bit depending on tissue and condition, so the genome annotations are only so accurate. But they're a good start. 
    
    genomeAnnotation.df = genomeAnnotation.df[-which(!genomeAnnotation.df$type %in% c("exon")), ]
    
    annot = strsplit(x = genomeAnnotation.df$annot, ";", fixed = TRUE)
    
    genomeAnnotation.df$ID = gsub(pattern = "gene_id ", replacement = "", unlist(lapply(annot, "[[", 1)))
    
    
    # TSS for minus strand is NOT identical as if it was the plus strand
    # See http://www.researchgate.net/post/Can_someone_orient_me_to_the_coordinates_numbering_between_positive_and_negative_strand2
    
    gr.list = list()
    
    for (strandCur in c("+", "-")) {
      
      genomeAnnotationCur.df = genomeAnnotation.df[which(genomeAnnotation.df$strand == strandCur), ]
      
      if (strandCur == "+") {
        genomeAnnotationCur.df = genomeAnnotationCur.df[order(genomeAnnotationCur.df$ID, genomeAnnotationCur.df$start), ]
      } else {
        genomeAnnotationCur.df = genomeAnnotationCur.df[order(genomeAnnotationCur.df$ID, -genomeAnnotationCur.df$end), ]
      }
      
      # Find duplicate rows given the gene ID, quicker solution
      genomeAnnotationRed = genomeAnnotationCur.df$ID  
      isDuplicate = duplicated(genomeAnnotationRed) 
      genomeAnnotationRed.df =   genomeAnnotationCur.df[!isDuplicate, ]
      
      if (strandCur == "+") {
        gr.list[[strandCur]] <- GRanges(seqnames = Rle(genomeAnnotationRed.df$chr), 
                                        strand = genomeAnnotationRed.df$strand, 
                                        ranges = IRanges(start = genomeAnnotationRed.df$start, 
                                                         end = genomeAnnotationRed.df$start, 
                                                         names = genomeAnnotationRed.df$ID), 
                                        annot = genomeAnnotationRed.df$annot)
        
      } else {
        gr.list[[strandCur]] <- GRanges(seqnames = Rle(genomeAnnotationRed.df$chr), 
                                        strand = genomeAnnotationRed.df$strand, 
                                        ranges = IRanges(start = genomeAnnotationRed.df$end, 
                                                         end = genomeAnnotationRed.df$end, 
                                                         names = genomeAnnotationRed.df$ID), 
                                        annot = genomeAnnotationRed.df$annot)
        
      }
      
      
      
    }
    
    # https://support.bioconductor.org/p/93347/#93373
  

    gr = c(gr.list[[1]], gr.list[[2]])
    gr <- sort(sortSeqlevels(gr))
    
    flog.info(paste0("Found", length(gr), " TSS"))
    
    # If TSS are too close together, the calculation might be biased. Count only reads that uniquely map to one TSS 
    # Exclude TSS that are to close together, delete TSs so that all TSS have at least a distance of 2*thresholdWithinTSS
    
    # Compute the distance to the neighboring element
    res2 = distance(gr[-length(gr)], gr[-1])
    indexes = which(res2 < (thresholdWithinTSS + thresholdOutsideTSS))
    
    # Delete the i-th and(i+1)th element of indexes for each element
    deleteRows = sort(unique(c(indexes, (indexes + 1))))
    flog.info(paste0("Deleting ", length(deleteRows), " rows because they are located within ", (thresholdWithinTSS + thresholdOutsideTSS), " bp"))
    
    gr_red = gr[-deleteRows]
    
    write.table(paste(as.character(seqnames(gr_red)), 
                      as.character(ranges(gr_red)@start), 
                      as.character(ranges(gr_red)@start), 
                      names( gr_red), 
                      mcols(gr_red)$annot, 
                      as.character(strand(gr_red)), 
                      collapse = "\n", sep = "\t"), 
                file = filename_TSS, row.names = FALSE, quote = FALSE, col.names = FALSE, sep = "\t")
    
  } 
  
  TSS.df = read.table(filename_TSS, sep = "\t")
  colnames(TSS.df) = c("chr", "start", "end", "ID", "annotation", "strand")
  
  
  if (length(geneTypesToKeep) > 0 & length(which(geneTypesToKeep == "all")) == 0) {
    flog.info(paste0("Use only the following gene types and discard all others:", paste(geneTypesToKeep, collapse = ", ")))
    TSS.df =  TSS.df[which(grepl(paste(geneTypesToKeep, collapse = "|"), TSS.df$annotation)), ]
  }
  
  TSS.GRanges = GRanges(seqnames = Rle(TSS.df$chr), ranges = IRanges(start = TSS.df$start, end = TSS.df$end), strand = TSS.df$strand, annotation = TSS.df$ID)
  
  # Compile a list of positions that are located within the vicinity of TSS and count the reads that are located within
  TSS_within = promoters(TSS.GRanges, upstream = thresholdWithinTSS/2, downstream = thresholdWithinTSS/2)
  
  # Compile a list of positions that are located outside the vicinity of TSS and count the reads that are located within
  TSS_outside1 = flank(TSS_within, width = thresholdOutsideTSS, start = TRUE)
  #TSS_outside2 = flank(TSS_within, width=thresholdWithinTSS/2, start= FALSE)
  
  #TSS_outside = c(TSS_outside1, TSS_outside2)
  
  param <- ScanBamParam(which = TSS_within, flag = readFlags)  
  counts_within = countBam(BAMFile, param = param)
  
  param <- ScanBamParam(which = TSS_outside1, flag = readFlags)  
  counts_outside1 = countBam(BAMFile, param = param)
  
  #param <- ScanBamParam(which=TSS_outside2, flag = readFlags)  
  #counts_outside2 = countBam(BAMFile, param=param)
  
  # How many reads are inside a thresholdWithinTSS/2 bp(to either side) window of the TSS?
  nReads_within = sum(counts_within$records)
  nReads_outside = sum(counts_outside1$records) #+ sum(counts_outside2$records)
  
  enrichment = nReads_within / nReads_outside
  
  fractionWindowSizes = thresholdWithinTSS / thresholdOutsideTSS
  enrichmentCorrected =  enrichment / fractionWindowSizes
  
  
  counts_within = scanBam(BAMFile, param = ScanBamParam(what = c("isize")))
  
  
  
  flog.info(paste("Found ", nReads_within, " reads within [-", 
                  thresholdWithinTSS/2, ":+", thresholdWithinTSS/2, "] bp of annotated TSS and ", 
                  nReads_outside, " reads in(-", 
                  ((thresholdWithinTSS/2) + thresholdOutsideTSS), 
                  ":-", thresholdWithinTSS/2, #" and +", 
                  #thresholdWithinTSS/2, ":+", ((thresholdWithinTSS/2) + thresholdOutsideTSS/2), 
                  "). ", "Overall TSS enrichment: ", 
                  round(enrichment, 2), "(corrected:", 
                  round(enrichmentCorrected, 2), 
                  ")", sep = " "))
  
  return(list("countsAroundTSS" = counts_within, 
              "countsUpstreamTSS" = counts_outside1, 
              "overallEnrichmentRaw" = round(enrichment, 2), 
              "overallEnrichmentNorm" = round(enrichmentCorrected, 2), 
              #"countsDownstreamTSS" = counts_outside2, 
              "par" = list(
                "gencodeVersion" = annotationFile, 
                "BAMFile" = BAMFile, 
                "thresholdWithinTSS" = thresholdWithinTSS, 
                "thresholdOutsideTSS" = thresholdOutsideTSS, 
                "readFlags" = readFlags
                
              )
              
  ))
  
}


splitStringInMultipleLines <- function(input, width, sepChar = "\n", verbose = TRUE) {
    
    assertCharacter(input, min.len = 1)
    assertInt(width, lower = 1)
    assertCharacter(sepChar, len = 1)
    
    
    as.character(sapply(input, function(x) {
        paste0(strsplit(x, paste0("(?<=.{", width, "})"), perl = TRUE)[[1]], collapse = sepChar)
    }
    ))
    
}


doPCAPlot <- function(file_readCounts, file_metadata, file_output, file_log, metadataToInclude, type, 
                      binSize = 10000, filterChr = TRUE, splitByGCCorrection = TRUE, minNoCountsRows = 0) {
  
  
  start.time  <-  Sys.time()
  par.l = list()
  par.l$log_minlevel = "INFO"

  checkAndLoadPackages(c("tidyverse", "futile.logger",  "checkmate", "tools", "methods", "DESeq2", "readr"), verbose = TRUE)
  
  
  
  coldata = read_tsv(file_metadata)
  
  if (nrow(problems(coldata)) > 0) {
      flog.fatal(paste0("Parsing errors: "), problems(coldata), capture = TRUE)
      stop("Error when parsing the file ", file_metadata, ", see errors above")
  }
  
  startLogger(file_log, par.l$log_minlevel, removeOldLog = TRUE)
  
  countsPCA = read_tsv(file_readCounts)
  if (nrow(problems(countsPCA)) > 0) {
    flog.fatal(paste0("Parsing errors: "), problems(countsPCA), capture = TRUE)
    stop("Error when parsing the file ", file_readCounts, ", see errors above")
  }
 
  # Test purposes
  #countsPCA =  countsPCA[1:1000,]
  
  colnames(countsPCA)[1] = "chr"
  
  if (filterChr) {
    countsPCA = dplyr::filter(countsPCA, !grepl("random|chrUn|chrM|hap|_gl", chr, perl = TRUE))
  }
  
  
  if (type == "all") {
      colnames(countsPCA) = gsub(".final.bam", "", colnames(countsPCA))
     
  } else if (type == "pooled") {
      colnames(countsPCA) = gsub(".merged", "", colnames(countsPCA))
      colnames(countsPCA) = gsub(".final.bam", "", colnames(countsPCA))
      
      # Rewrite coldata
      coldata = coldata[duplicated(coldata[,"individual"]),]
      coldata$sampleName = coldata$individual
      
  } else {
      flog.fatal(paste0("Unsupported type ", type))
  }

  colnames(countsPCA) = gsub("'", "", colnames(countsPCA))
  
  
  
  allMatrices.l = list()
  allColdata.l = list()
  
  if (splitByGCCorrection) {
    
    countsPCA_original = dplyr::select(countsPCA, -contains("noGCBias"))
    allMatrices.l[["original"]] = as.matrix(countsPCA_original[,-c(1:3)])
    allColdata.l [["original"]] = coldata
    
    countsPCA_GC = dplyr::select(countsPCA, contains("noGCBias"))
    allMatrices.l[["GCCorrected"]] = as.matrix(countsPCA_GC)
    coldata_GC = coldata
    coldata_GC$sampleName = paste0(coldata_GC$sampleName, ".noGCBias")
    coldata_GC$individual = paste0(coldata_GC$individual, ".noGCBias")
    allColdata.l [["GCCorrected"]] = coldata_GC
    
    allMatrices.l[["all"]] = as.matrix(countsPCA[,-c(1:3)])
    coldata_all = rbind(coldata, coldata_GC)
    allColdata.l [["all"]] = coldata_all
    
  } else {
    
    allMatrices.l[["all"]] = countsPCA
  }
  
  
  pdf(file_output)
  
  for (matrixCur in names(allMatrices.l)) {
   
   flog.info(paste0("Generate plots for group ", matrixCur))
   
   matrixCur.m = allMatrices.l[[matrixCur]]
   colDataCur  = allColdata.l[[matrixCur]]

   assertSubset(colnames(matrixCur.m), colDataCur$sampleName)
   # Sort matrix to ensure same order as coldata table
   matrixCur.m = matrixCur.m[, colDataCur$sampleName]
   
   # Use DESeq2 for it
   dds <- DESeqDataSetFromMatrix(countData = matrixCur.m, colData = colDataCur, design = ~1)
   
   if (minNoCountsRows > 0) {
     dds <- dds[ rowSums(counts(dds)) > minNoCountsRows, ]
   }
  
   vsd <- varianceStabilizingTransformation(dds, blind = TRUE)
   
   for (nrowsCur in c(500,5000,50000, nrow(vsd))) {
     
     for (varCur in metadataToInclude) {
       pcadata = plotPCA(vsd, intgroup = varCur, ntop = nrowsCur, returnData = TRUE)
       
       percentVar <- round(100 * attr(pcadata, "percentVar"))
       g = ggplot(pcadata, aes_string("PC1", "PC2", color = varCur)) +
         geom_point(size = 2) +
         xlab(paste0("PC1: ",percentVar[1],"% variance")) +
         ylab(paste0("PC2: ",percentVar[2],"% variance")) + 
         coord_fixed() + theme_bw() + 
         ggtitle(paste0("Group ", matrixCur, "\nTop ", .prettyNum(nrowsCur), " ", binSize, " bp bins")) + 
         theme(plot.title = element_text(hjust = 0.5))
       plot(g)
       
     }
     
   }
  }

  dev.off()
  
  

}
