
#TODO: Set 0 counts to 0.1 or 1 to allow enrichment values
#Soll man paired end reads doppelt zählen? Strange effect with even and odd numbers


# modified from: https://github.com/imbforge/encodeChIPqc/blob/master/R/PBC.R
#' PCR bottleneck coefficient
#'
#' Calculate the PCR bottleneck coefficient as described in the ENCODE
#' guidelines.
#'
#' The PCR bottleneck coefficient (PBC) is a measure of library complexity, i.e. how skewed the
#' distribution of read counts per location is towards 1 read per location.
#' 
#' Defined in the ENCODE guidelines (https://genome.ucsc.edu/ENCODE/qualityMetrics.html) as:
#'
#' PBC = N1/Nd
#' 
#' with
#' \itemize{
#' \item{\code{N1}: Number of genomic locations to which EXACTLY one unique mapping read maps.}
#' \item{\code{Nd}: Number of genomic locations to which AT LEAST one unique mapping read maps, i.e.
#' the number of non-redundant, unique mapping reads.}
#' }
#' 
#' PBC is further described on the ENCODE Software Tools page. Provisionally, 0-0.5 is severe
#' bottlenecking, 0.5-0.8 is moderate bottlenecking, 0.8-0.9 is mild bottlenecking, while 0.9-1.0
#' is no bottlenecking. Very low values can indicate a technical problem, such as PCR bias, or a
#' biological finding, such as a very rare genomic feature. Nuclease-based assays (DNase, MNase)
#' detecting features with base-pair resolution (transcription factor footprints, positioned 
#' nucleosomes) are expected to recover the same read multiple times, resulting in a lower PBC
#' score for these assays. Note that the most complex library, random DNA, would approach 1.0, 
#' thus the very highest values can indicate technical problems with libraries. It is the practice
#' for some labs outside of ENCODE to remove redundant reads; after this has been done, the value
#' for this metric is 1.0, and this metric is not meaningful. 82\% of TF ChIP, 89\% of His ChIP, 77\%
#' of DNase, 98\% of FAIRE, and 97\% of control ENCODE datasets have no or mild bottlenecking.
#'
#' @param The path to the \code{.bam} file of a ChIP sample or a \code{GAlignments} object of the ChIP sample.
#' @return The PBC coefficient.
#' @examples
#' pbc <- PBC("IP.bam")
calcPBCValues <- function(BAMFile, verbose = TRUE) {
  
  require(GenomicAlignments)
  require(data.table)
  
  assertFlag(verbose)
  assert(checkFile(BAMFile, access = "r"), checkClass(BAMFile, "GAlignments"))
  
  aln <- BAMFile

  if (class(BAMFile) != "GAlignments") {
    aln <- readGAlignments(BAMFile)
  }
  
  
  # convert GAlignments object to data.table for fast aggregation
  aln <- data.table(
    strand=as.factor(BiocGenerics::as.vector(strand(aln))),
    seqnames=as.factor(BiocGenerics::as.vector(seqnames(aln))),
    pos=ifelse(strand(aln) == "+", start(aln), end(aln))
  )
  
  # aggregate reads by position and count them
  readsPerPosition <- aln[,list(count=.N), by=list(strand, seqnames, pos)]$count
  
  # PBC1 = positions with exactly 1 read / positions with at least 1 read
  m1 = sum(readsPerPosition == 1)
  mDistinct =length(readsPerPosition)
  PBC1 <-  m1 / mDistinct
  
  # PBC2 = positions with exactly 1 read / positions with exactly 2 reads
  m2 = sum(readsPerPosition == 2)
  PBC2 <-  m1 / m2
  

  res.l = list("PBC1" = list(
                  "PBC1" = PBC1,
                  "M1" = m1,
                  "M_DISTINCT" = mDistinct
                ),
               "PBC2" = list(
                 "PBC2" = PBC2,
                 "M1" = m1,
                 "M2" = m2
               ))
  
  return(res.l)
}


annotationFile = "/g/scb2/zaugg/zaugg_shared/annotations/hg19/Gencode_v19/gencode.v19.annotation.gtf"
BAMFile = "/g/scb2/zaugg/carnold/Projects/AtacSeq/example/output/8.FinalOutput/test1_rep1.final.s.bam"
# checkEnrichmentOverTSS(annotationFile, BAMFile, plotDistrFile = "test.pdf") 


calcLibraryComplexity <- function(BAMFile, experimentType = "ATAC-Seq", verbose = TRUE) {
  
  assertFile(BAMFile, access = "r")
  checkSubset(experimentType, c("ATAC-Seq", "ChIP-Seq"))
  assertFlag(verbose) 
  
  flog.info(paste0("Calculate the following measures for file ", BAMFile, ":")) 
  flog.info(paste0("1) Non-Redundant Fraction (NRF) - Number of distinct uniquely mapping reads (i.e. after removing duplicates) / Total number of reads. ")) 
  flog.info(paste0("2) PCR Bottlenecking Coefficient 1 (PBC1): PBC1= M1 / M_DISTINCT")) 
  flog.info(paste0("3) PCR Bottlenecking Coefficient 2 (PBC2): PBC2= M1 / M2")) 
  flog.info(paste0(" M1: number of genomic locations where exactly one read maps uniquely")) 
  flog.info(paste0(" M_DISTINCT: number of distinct genomic locations to which some read maps uniquely")) 
  flog.info(paste0(" M2: number of genomic locations where two reads map uniquely")) 
  flog.info(paste0("The following evaluations are based on the numbers from https://www.encodeproject.org/data-standards/terms/#library, January 2017"))
  
  flog.info(paste0("Results for ", experimentType, " data:"))
  
  PBC.l = calcPBCValues(BAMFile, verbose = verbose) 
  PBC1 = PBC.l$PBC1$PBC1
  PBC2 = PBC.l$PBC2$PBC2
  
  
  # The following numbers are taken from https://www.encodeproject.org/data-standards/terms/#library, January 2017
  
  # Structure of list: two elements correspond to PBC1 (lower and upper limit, respectively)
  status_PBC1.l  = list(
    "ATAC-Seq" = list(
      "severe"   = c(0, 0.7),
      "moderate" = c(0.7, 0.9),
      "none"      = c(0.9, 1)
    ),
    "ChIP-Seq" = list(
      "severe"   = c(0, 0.5),
      "moderate" = c(0.5, 0.8),
      "mild"     = c(0.8, 0.9),
      "none"      = c(0.9, 1)
    )
    
  )
  
  statusPBC1 = NA
  for (statusCur in names(status_PBC1.l[[experimentType]])) {
    if (PBC1 >= status_PBC1.l[[experimentType]][[statusCur]][1] && PBC1 <= status_PBC1.l[[experimentType]][[statusCur]][2]) {
          statusPBC1 = statusCur
    }
  }

  
  flog.info(paste0(" PBC1=", round(PBC1,2), ", bottlenecking level \"", statusPBC1, 
                   "\" (bottlenecking levels: ", paste0(names(status_PBC1.l[[experimentType]]), collapse = ", "), ")"))
  
  
  # The following numbers are taken from https://www.encodeproject.org/data-standards/terms/#library, January 2017
  
  # Structure of list: two elements correspond to PBC2 (lower and upper limit, respectively)
  
  status_PBC2.l  = list(
    "ATAC-Seq" = list(
      "severe"   = c(0, 1),
      "moderate" = c(1, 3),
      "none"      = c(3, 1000000)
    ),
    "ChIP-Seq" = list(
      "severe"   = c(0, 1),
      "moderate" = c(1, 3),
      "mild"     = c(3, 10),
      "none"      = c(10, 1000000)
    )
  )
    
  
  statusPBC2 = NA
  for (statusCur in names(status_PBC2.l[[experimentType]])) {
    if (PBC2 >= status_PBC2.l[[experimentType]][[statusCur]][1] && PBC2 <= status_PBC2.l[[experimentType]][[statusCur]][2]) {
      statusPBC2 = statusCur
    }
  }
  
  
  flog.info(paste0(" PBC2=", round(PBC2,2), ", bottlenecking level \"", statusPBC2, 
                   "\" (bottlenecking levels: ", paste0(names(status_PBC2.l[[experimentType]]), collapse = ", "), ")"))

  
  
  ###############
  ## NRF status #
  ###############
  
  NRF = 1
  #NRF = nReadsDistinctUnique / nReadsTotal
  
  # The following numbers are taken from https://www.encodeproject.org/data-standards/terms/#library, January 2017
  
  # Structure of list: two elements correspond to NRF (lower and upper limit, respectively)
  
  status_NRF.l  = list(
    "ATAC-Seq" = list(
      "concerning" = c(0, 0.7),
      "acceptable" = c(0.7, 0.9),
      "ideal"      = c(0.9, 1)
    ),
    "ChIP-Seq" = list(
      "concerning" = c(0, 0.5),
      "acceptable" = c(0.5, 0.8),
      "compliant"  = c(0.8, 0.9),
      "ideal"      = c(0.9, 1)
    )
  
  )
  
  statusNRF = "NA"
  for (statusCur in names(status_NRF.l[[experimentType]])) {
    if (NRF >= status_NRF.l[[experimentType]][[statusCur]][1] && NRF <= status_NRF.l[[experimentType]][[statusCur]][2]) {
      statusNRF = statusCur
    }
  }
  

  flog.info(paste0(" NRF=", NRF, ", library complexity \"", statusNRF, 
                   "\" (complexity levels: ", paste0(names(status_NRF.l[[experimentType]]), collapse = ", "), ")"))
  
  
  
}


checkEnrichmentOverTSS <- function(annotationFile, BAMFile, thresholdWithinTSS = 4000, thresholdOutsideTSS = 1000, 
                                   readFlags = NULL, geneTypesToKeep = c("protein_coding"), doAlsoEncodeVariant = TRUE, assemblyVersion = NULL, plotDistrFile = NULL, verbose = TRUE) {
  
  checkAndLoadPackages(c("tools", "GenomicRanges"), verbose = verbose)
  
  assertFile(annotationFile, access = "r")
  assertFile(BAMFile, access = "r")
  assertSubset(tolower(file_ext(BAMFile)), c("bam"))
  assertInt(thresholdWithinTSS, lower = 1)
  assertInt(thresholdOutsideTSS, lower = 1)
  assert(checkNull(readFlags), checkIntegerish(readFlags, len = 2, lower = 0))
  assert(checkNull(assemblyVersion), checkSubset(assemblyVersion, c("hg19", "hg38", "mm9", "mm10")))
  assertCharacter(geneTypesToKeep, min.chars = 1, min.len = 1)
  assertFlag(verbose) 
  assert(checkNull(plotDistrFile), checkCharacter(plotDistrFile, len = 1))
  if (!testNull(plotDistrFile)) {
   assertDirectoryExists(dirname(plotDistrFile), access = "w")
  }    
  
  
  if (testNull(readFlags)) {
    readFlags = scanBamFlag()
  }
  
  
  flog.info(paste0("Check enrichment over TSS for file: ", BAMFile))
  flog.info(paste("thresholdWithinTSS=", thresholdWithinTSS, ", thresholdOutsideTSS=", thresholdOutsideTSS))
  
  flog.info(paste("Retrieving TSS annotation from file", annotationFile, sep = " "))
  filename_TSS = paste0(annotationFile, ".TSS_nonOverlapping", (thresholdWithinTSS + thresholdOutsideTSS), ".bed")
  
  if (!file.exists(filename_TSS)) {
    
    flog.info(paste0("Could not find file", filename_TSS, ", calculating TSS anew...", sep = " "))
    
    # Read only particular columns(positions and type) 
    colsToRead = c("character", "NULL", "character", "integer", "integer", "NULL", "character", "NULL", "character")
    
    genomeAnnotation.df =  read.table(annotationFile, colClasses = colsToRead, skip = 5, sep = "\t")
    colnames( genomeAnnotation.df) = c("chr", "type", "start", "end", "strand", "annot")
    
    # How to define a TSS? Current solution:
    # The first codon of the first exon is approximately the transcription start site. The translation start site(i.e. the start of the CDS) is located somewhere within the transcript, but doesn't necessarily have to be in the first exon at all.
    # The first nucleotide of the first exon is a good approximation. The main issue with it is that the exact TSS can shift around a bit depending on tissue and condition, so the genome annotations are only so accurate. But they're a good start. 
    
    genomeAnnotation.df = genomeAnnotation.df[-which(!genomeAnnotation.df$type %in% c("exon")), ]
    
    annot = strsplit(x = genomeAnnotation.df$annot, ";", fixed = TRUE)
    
    genomeAnnotation.df$ID = gsub(pattern = "gene_id ", replacement = "", unlist(lapply(annot, "[[", 1)))
    
    
    # TSS for minus strand is NOT identical as if it was the plus strand
    # See http://www.researchgate.net/post/Can_someone_orient_me_to_the_coordinates_numbering_between_positive_and_negative_strand2
    
    gr.list = list()
    
    for (strandCur in c("+", "-")) {
      
      genomeAnnotationCur.df = genomeAnnotation.df[which(genomeAnnotation.df$strand == strandCur), ]
      
      if (strandCur == "+") {
        genomeAnnotationCur.df = genomeAnnotationCur.df[order(genomeAnnotationCur.df$ID, genomeAnnotationCur.df$start), ]
      } else {
        genomeAnnotationCur.df = genomeAnnotationCur.df[order(genomeAnnotationCur.df$ID, -genomeAnnotationCur.df$end), ]
      }
      
      # Find duplicate rows given the gene ID, quicker solution
      genomeAnnotationRed = genomeAnnotationCur.df$ID  
      isDuplicate = duplicated(genomeAnnotationRed) 
      genomeAnnotationRed.df =   genomeAnnotationCur.df[!isDuplicate, ]
      
      if (strandCur == "+") {
        gr.list[[strandCur]] <- GRanges(seqnames = Rle(genomeAnnotationRed.df$chr), 
                                        strand = genomeAnnotationRed.df$strand, 
                                        ranges = IRanges(start = genomeAnnotationRed.df$start, 
                                                         end = genomeAnnotationRed.df$start, 
                                                         names = genomeAnnotationRed.df$ID), 
                                        annot = genomeAnnotationRed.df$annot)
        
      } else {
        gr.list[[strandCur]] <- GRanges(seqnames = Rle(genomeAnnotationRed.df$chr), 
                                        strand = genomeAnnotationRed.df$strand, 
                                        ranges = IRanges(start = genomeAnnotationRed.df$end, 
                                                         end = genomeAnnotationRed.df$end, 
                                                         names = genomeAnnotationRed.df$ID), 
                                        annot = genomeAnnotationRed.df$annot)
        
      }
      
      
      
    }
    
    gr = c(gr.list[[1]], gr.list[[2]])
    gr <- sort(sortSeqlevels(gr))
    
    flog.info(paste0("Found", length(gr), " TSS"))
    
    # If TSS are too close together, the calculation might be biased. Count only reads that uniquely map to one TSS 
    # Exclude TSS that are to close together, delete TSs so that all TSS have at least a distance of 2*thresholdWithinTSS
    
    # Compute the distance to the neighboring element
    res2 = distance(gr[-length(gr)], gr[-1])
    indexes = which(res2 < (thresholdWithinTSS + thresholdOutsideTSS))
    
    # Delete the i-th and(i+1)th element of indexes for each element
    deleteRows = sort(unique(c(indexes, (indexes + 1))))
    flog.info(paste0("Deleting ", length(deleteRows), " rows because they are located within ", (thresholdWithinTSS + thresholdOutsideTSS), " bp"))
    
    gr_red = gr[-deleteRows]
    
    write.table(paste(as.character(seqnames(gr_red)), 
                      as.character(ranges(gr_red)@start), 
                      as.character(ranges(gr_red)@start), 
                      names( gr_red), 
                      mcols(gr_red)$annot, 
                      as.character(strand(gr_red)), 
                      collapse = "\n", sep = "\t"), 
                file = filename_TSS, row.names = FALSE, quote = FALSE, col.names = FALSE, sep = "\t")
    
  } 
  
  TSS.df = read.table(filename_TSS, sep = "\t")
  colnames(TSS.df) = c("chr", "start", "end", "ID", "annotation", "strand")
  
  
  if (length(geneTypesToKeep) > 0 & length(which(geneTypesToKeep == "all")) == 0) {
    flog.info(paste0("Use only the following gene types and discard all others:", paste(geneTypesToKeep, collapse = ", ")))
    TSS.df =  TSS.df[which(grepl(paste(geneTypesToKeep, collapse = "|"), TSS.df$annotation)), ]
  }
  
  TSS.GRanges = GRanges(seqnames = Rle(TSS.df$chr), ranges = IRanges(start = TSS.df$start, end = TSS.df$end), strand = TSS.df$strand, annotation = TSS.df$ID)
  
  # Compile a list of positions that are located within the vicinity of TSS and count the reads that are located within
  TSS_within = promoters(TSS.GRanges, upstream = thresholdWithinTSS/2, downstream = thresholdWithinTSS/2)
  
  # Compile a list of positions that are located outside the vicinity of TSS and count the reads that are located within
  TSS_outside1 = flank(TSS_within, width = thresholdOutsideTSS, start = TRUE)
  #TSS_outside2 = flank(TSS_within, width=thresholdWithinTSS/2, start= FALSE)
  #TSS_outside = c(TSS_outside1, TSS_outside2)
  

  
  
  param <- ScanBamParam(which = TSS_within, flag = readFlags)  
  counts_within = countBam(BAMFile, param = param)
  
  param <- ScanBamParam(which = TSS_within, what = c("rname", "strand", "pos", "qwidth"), flag = readFlags) 
  reads_within  = scanBam(BAMFile, param = param)
  
  bamCoverage <- CoverageBamFile(BAMFile, run_type = "paired")
  
  binWidth = 50
  normalizationProcedure = "rpm"
  normalizationProcedure = NULL
  
  # Create a bed file
  filename_TSS = "TSS.bed"
  export.bed(TSS.GRanges[1:500,], con = filename_TSS)
  
  cov.m  = cov.matrix(bamCoverage, coordfile = filename_TSS, normalization = normalizationProcedure, 
                      extend = 2000, num_cores = 4, bin_width = binWidth)
  
  # Does not seem to work properly...
  draw.heatmap(cov.m,outfile="TSS_profile.png")
  
  library(GenomicAlignments)
  galign = readGAlignmentPairs(BAMFile, param = param)
  reads_coverage = GenomicAlignments::coverage(galign)
  
  reads_coverage1 = GenomicAlignments::coverage(BamFile(BAMFile, asMates=TRUE), param = param)
  reads_coverage2 = GenomicAlignments::coverage(BamFile(BAMFile, asMates=TRUE))
  
  # Delete chromosomes we dont care about, also prevents an error message...
  reads_coverage1 = reads_coverage1[1:24]
  
  aggregate(reads_coverage1, TSS.GRanges[1:5,], FUN = mean)
  
  param <- ScanBamParam(which = TSS_outside1, flag = readFlags)  
  counts_outside1 = countBam(BAMFile, param = param)
  
  
  oo <- order(as.factor(seqnames(TSS.GRanges[1:500,])))
  myRegions <- TSS.GRanges[1:500,][oo]
  
  GRangesList(myRegions)
  
  res = Views(reads_coverage1, as(myRegions,"RangesList"))
  
  
  pparam <- PileupParam(distinguish_nucleotides=FALSE, distinguish_strands=TRUE)
  res = pileup(BamFile(BAMFile, asMates=TRUE), scanBamParam=param, pileupParam=pparam)
  
  
  #param <- ScanBamParam(which=TSS_outside2, flag = readFlags)  
  #counts_outside2 = countBam(BAMFile, param=param)
  
  if (doAlsoEncodeVariant) {
    
    
    thresholdWithinTSSTotalEncode = 2000
    TSS_withinEncode = promoters(TSS.GRanges, upstream = thresholdWithinTSSTotalEncode/2, downstream = thresholdWithinTSSTotalEncode/2)
    
    thresholdOutsideTSSTotalEncode = 100
    TSS_outsideEncode1 = flank(TSS_withinEncode, width = thresholdOutsideTSSTotalEncode, start= FALSE, both = FALSE)
    TSS_outsideEncode2 = flank(TSS_withinEncode, width = thresholdOutsideTSSTotalEncode, start= TRUE , both = FALSE)
    
    param <- ScanBamParam(which = TSS_withinEncode, flag = readFlags)  
    counts_withinEncode = countBam(BAMFile, param = param)
    
    param <- ScanBamParam(which = TSS_outsideEncode1, flag = readFlags)  
    counts_outsideEncode1 = countBam(BAMFile, param = param)
  
    param <- ScanBamParam(which = TSS_outsideEncode2, flag = readFlags)  
    counts_outsideEncode2 = countBam(BAMFile, param = param)
    
    # Counts reads outside and calculate fold change for inside
    
    # Calculate average coverage in the 100bp flanking regions
    
    # Fold change based on the average read depth from the flanking regions
    
    # value at the center, the TSS, as signal value
    
    # 
    
    
    # TSS enrichment result evaluation, based on https://www.encodeproject.org/data-standards/atac-seq/
    
    status_TSS_Enrichment.l  = list(
            "hg19" = list(
              "concerning" = c(0, 6),
              "acceptable" = c(6, 10),
              "ideal"      = c(10,1000000)
            ),
            "hg38" = list(
              "concerning" = c(0, 5),
              "acceptable" = c(5, 7),
              "ideal"      = c(7,1000000)
            ),
            "mm9" = list(
              "concerning" = c(0, 5),
              "acceptable" = c(5, 7),
              "ideal"      = c(7,1000000)
            ),
            "mm10" = list(
              "concerning" = c(0, 10),
              "acceptable" = c(10, 15),
              "ideal"      = c(15,1000000)
            )

    )
    
    assertSubset(assemblyVersion, names(status_TSS_Enrichment.l))
    
    
    status = "NA"
    for (statusCur in names(status_TSS_Enrichment.l[[assemblyVersion]])) {
      if (enrichment >= status_TSS_Enrichment.l[[assemblyVersion]][[statusCur]][1] && enrichment <= status_TSS_Enrichment.l[[assemblyVersion]][[statusCur]][2]) {
        status = statusCur
      }
    }
    
    
    
  }

  
  # How many reads are inside a thresholdWithinTSS/2 bp(to either side) window of the TSS?
  nReads_within = sum(counts_within$records)
  nReads_outside = sum(counts_outside1$records) #+ sum(counts_outside2$records)
  
  enrichment = nReads_within / nReads_outside
  
  fractionWindowSizes = thresholdWithinTSS / thresholdOutsideTSS
  enrichmentCorrected =  enrichment / fractionWindowSizes
  
  # TODO: WHy here?
  #counts_within = scanBam(BAMFile, param = ScanBamParam(what = c("isize")))
  
  
  
  flog.info(paste("Found ", nReads_within, " reads within [-", 
                   thresholdWithinTSS/2, ":+", thresholdWithinTSS/2, "] bp of annotated TSS and ", 
                   nReads_outside, " reads in(-", 
                   ((thresholdWithinTSS/2) + thresholdOutsideTSS), 
                   ":-", thresholdWithinTSS/2, #" and +", 
                   #thresholdWithinTSS/2, ":+", ((thresholdWithinTSS/2) + thresholdOutsideTSS/2), 
                   "). ", "Overall TSS enrichment: ", 
                   round(enrichment, 2), "(corrected:", 
                   round(enrichmentCorrected, 2), 
                   ")", sep = " "))
  
  return(list("countsAroundTSS" = counts_within, 
              "countsUpstreamTSS" = counts_outside1, 
              "overallEnrichmentRaw" = round(enrichment, 2), 
              "overallEnrichmentNorm" = round(enrichmentCorrected, 2), 
              #"countsDownstreamTSS" = counts_outside2, 
              "par" = list(
                "gencodeVersion" = annotationFile, 
                "BAMFile" = BAMFile, 
                "thresholdWithinTSS" = thresholdWithinTSS, 
                "thresholdOutsideTSS" = thresholdOutsideTSS, 
                "readFlags" = readFlags
                
              )
              
  ))
  
}